/***************************************************************   
IMSE1-C.PRG

written by
Bruce E. Hansen
Department of Economics
University of Wisconsin
www.ssc.wisc.edu/~bhansen
bhansen@ssc.wisc.edu

This Gauss program numerically calculates integrals I1 and I2 needed
for evaluation of MISE as described in the  paper
"Exact Mean Integrated Squared Error of Higher-Order Kernel Estimators"

The program loads the file "legendre.fmt" which contains points of evaluation
for Gauss-Legendre numerical integration.

The program creates the output "imse1.gcg" which is needed for the programs 
"imse1-p.prg" and "regret.prg"

This program takes several hours to run.

*****************************************************************/

let rs = 1 2 3 4 5 6 7 8 9 10;
rn=rows(rs);
let ss = 1 5 8;
sn=rows(ss);
_mmax=15;

loadm legendre;
q=4;
xx=packr(legendre[.,2*q-1]);
ww=packr(legendre[.,2*q]);
inc=.001;
hn=round(3/inc);	@ creates centered bandwidth sequence @
hs=seqa(inc,inc,hn);
let bbs=10 10 40 30 30 10 10 10 10 30 110 50 80 90 50;

/*****************************************************************/

int1=arrayalloc(_mmax|(sn+1)|rn|hn,0);
int2=arrayalloc(_mmax|(sn+1)|rn|hn,0);
intd1=zeros(_mmax,hn);
intt1=zeros(_mmax,hn);
intt2=zeros(_mmax,hn);

_m=1; do while _m<=_mmax;

bb=bbs[_m];
x=((-rev(xx))|xx)*(bb/2)+bb/2;
w=(rev(ww)|ww)*(bb/2);

{p,mu,sig} = models(_m);
fs=abs((expt(-((x*(sig')).^2)/2).*expt(x*sqrt(-1)*(mu')))*p).^2;
sd=mn_sd(_m);
s0=sqrt((sig.^2)+(sig.^2)');
x0=(mu-mu')./s0;
f0=expt(-(x0.^2)/2)/sqrt(2*pi)./s0;
i0=p'f0*p;

h0=TrapRuleThumb(50,sd);
h1=hs*h0;
i1=zeros(hn,1);
i2=zeros(hn,1);
for hi (1,hn,1);
  k=((x*h1[hi]).<=1); 
  intd1[_m,hi]=((k.*fs)'w)/pi;
endfor;

i1=zeros(hn,1);
i2=zeros(hn,1);
for hi (1,hn,1);
  k=TrapCF(x*h1[hi]);
  intt1[_m,hi]=((k.*fs)'w)/pi;
  intt2[_m,hi]=(((k.^2).*fs)'w)/pi;
endfor;

for ri (1,rn,1);
  r=rs[ri];
  h0=GaussRuleThumb(50,r,sd);
  h1=hs*h0;
  for hi (1,hn,1);
    k=GaussCF(x*h1[hi],r); 
    int1[_m,1,ri,hi]=((k.*fs)'w)/pi;
    int2[_m,1,ri,hi]=(((k.^2).*fs)'w)/pi;
  endfor;
endfor;

for ri (1,rn,1);
  r=rs[ri];
  si=1; do while si<=sn;
    s=ss[si];
    h0=PolyRuleThumb(50,r,s,sd);
    h1=hs*h0;
    for hi (1,hn,1);
      k=PolyCF(x*h1[hi],r,s); 
      int1[_m,si+1,ri,hi]=((k.*fs)'w)/pi;
      int2[_m,si+1,ri,hi]=(((k.^2).*fs)'w)/pi;
    endfor;
  si=si+1;endo;
endfor;
_m=_m+1;endo;

saveall imse1;

/*****************************************************/


@ Gaussian Kernel Roughness @
proc GaussRoughness(r);
local j,a,rg;
  j=seqa(0,1,r);
  a=1;
  if r>1;a=a|(1./cumprodc(seqa(2,2,r-1)));
  endif;
  rg=(a'gamma(j+j'+.5)*a)/2/pi;
retp(rg);
endp;

@ Rule-of-Thumb for Gaussian kernels @
proc GaussRuleThumb(n,r,sd);
local h,c,rf,m2;
  m2=(2^(2*r))*(gamma(r+.5)^2)/pi;
  rf=gamma(2*r+.5)/2/pi;
  c=GaussRoughness(r)/rf/m2/4/r*(((2*r)!).^2);
  h=(c/n)^(1/(1+4*r));
  h=h*sd;
retp(h);
endp;

@ Characteristic Function of Gaussian Kernel @
proc GaussCF(x,r);
local t,tt,f,ff;
    t=vecr(x);
    tt=((t.^2)/2);
    f=expt(-tt);
    if r>1;
      f=f.*(1+sumc(cumprodc((1./seqa(1,1,r-1)).*(tt'))));
    endif;
    ff=reshape(f,rows(x),cols(x));
retp(ff);
endp;

/******************************************************/

proc PolyRoughness(r,s);
local j,a,jk,jkf,b,rg,jk1,bb;
  a=1;
  if r>1;
    a=a|cumprodc((s-.5)./seqa(1,1,r-1)+1);
  endif;
  a=(a.*seqa(s+.5,2,r))*gamma(s+.5);
  j=seqa(0,1,r);
  jk=(j-j'+s);
  jk1=(jk.>=0);
  jkf=(jk.*jk1)!;
  bb=gamma(j+j'+2*s+1.5)./gamma(j+j'+.5);
  b=(jk1.*(jk1'))./(jkf.*(jkf'))./bb; 
  rg=(a'b*a)*((2*s)!)/pi;
retp(rg);
endp;


@ Rule-of-Thumb for Polynomial kernels @
proc PolyRuleThumb(n,r,s,sd);
local h,c,rf,m2;
  m2=((gamma(r+s+.5)*gamma(r+.5)/gamma(2*r+s+.5)).^2)/pi;
  rf=gamma(2*r+.5)/2/pi;
  c=PolyRoughness(r,s)/rf/m2/4/r*(((2*r)!).^2);
  h=(c/n)^(1/(1+4*r));
  h=h*sd;
retp(h);
endp;


@ Characteristic Function of Polynomial Kernel @
proc PolyCF(x,r,s);
local tk,tt,f,ti,b,a,t,ff;
    tk=vecr(x);
    t=abs(tk);
    ti=(t.<= 1e-100);
    tt=t+ti;
    a=1;
    if r>1;
      a=a|cumprodc((s-.5)./seqa(1,1,r-1)+1);
    endif;
    a=(a.*seqa(s+.5,2,r))*gamma(s+.5);
    b=SphericalBessel(tt,s+2*r-2);
    f=b[.,seqa(s,2,r)]*a;
    f=f.*((2./tt).^s);
    f=f*(2/sqrt(pi));
    f=f.*(1-ti)+ti;    
    f=missrv(f,1);
    ff=reshape(f,rows(x),cols(x));
retp(ff);
endp;

/******************************************************/


@ Trapezoidal Kernel  @
fn TrapRuleThumb(n,sd)=sd/sqrt(ln((n-1)));
fn TrapCF(t)=(2-t).*(t.<=2)-(1-t).*(t.<=1);


/******************************************************/


@ Standard deviation of Mixture-Normal Densities @
proc mn_sd(_m);
local p,mu,sig,m,m2,s;
  {p,mu,sig} = models(_m);
  m=p'mu;
  m2=p'((sig.^2)+mu.^2);
  s=sqrt(m2-m.^2);
retp(s);
endp;

@ Mixture-Normal Characteristic Function @
proc MNCF(_m,x);
local p,mu,sig,f1,f2,f,i;
  {p,mu,sig} = models(_m);
  i=sqrt(-1);
  f=(expt(-((x*(sig')).^2)/2).*expt(x*i*(mu')))*p;
retp(f);
endp;


@ Marron-Wand Mixture-Normal Parameterization @
proc (3) = models(_m);
local p,mu,sig;
if _m==1;
  p=1;mu=0; sig=1;
elseif _m==2;
  p= .2|.2|.6;
  mu= 0|.5|13/12;
  sig = 1|2/3|5/9;
elseif _m==3;
  p=ones(8,1)*(1/8);
  sig=(2/3).^seqa(0,1,8);
  mu=(sig-1)*3;
elseif _m==4;
  p=(2/3)|(1/3);
  mu=0|0;
  sig=1|.1;
elseif _m==5;
  p=.1|.9;
  mu=0|0;
  sig=1|.1;
elseif _m==6;
  p=.5|.5;
  mu=-1|1;
  sig=2/3|2/3;
elseif _m==7;
  p=.5|.5;
  mu=-1.5|1.5;
  sig=1/2|1/2;
elseif _m==8;
  p=.75|.25;
  mu=0|1.5;
  sig=1|1/3;
elseif _m==9;
  p=9/20|9/20|.1;
  mu=-1.2|1.2|0;
  sig=.6|.6|.25;
elseif _m==10;
  p=.5|.1|.1|.1|.1|.1;
  mu=0|seqa(-1,.5,5);
  sig=1|.1|.1|.1|.1|.1;
elseif _m==11;
  p=.49|.49|(ones(7,1)/350);
  mu=-1|1|seqa(-1.5,.5,7);
  sig=(2/3)|(2/3)|(ones(7,1)/100);
elseif _m==12;
  p=.5|((2.^seqa(3,-1,5))/31);
  mu=0|seqa(-1.5,1,5);
  sig=1|((2.^seqa(2,-1,5))/10);
elseif _m==13;
  p=.46|.46|(1/300)|(1/300)|(1/300)|(7/300)|(7/300)|(7/300);
  mu=-1|1|-.5|-1|-1.5|.5|1|1.5;
  sig=(2/3)|(2/3)|.01|.01|.01|.07|.07|.07;
elseif _m==14;
  p=(2.^seqa(5,-1,6))/63;
  mu=(65-(.5.^seqa(0,1,6))*96)/21;
  sig=(.5.^seqa(0,1,6))*32/63;
elseif _m==15;
  p=(2/7)|(2/7)|(2/7)|(1/21)|(1/21)|(1/21);
  mu=-(15/7)|-(3/7)|(9/7)|(16/7)|(18/7)|(20/7);
  sig=p;
endif;
retp(p,mu,sig);
endp;

/******************************************************/

@ trimmed exponential : Prevents Underflow @
proc expt(x);
local x1,x2,f;
  x1=(x.< -700);
  x2=(x.> 700);
  f=exp(x.*(1-x1-x2))-x1+x2*exp(700);
retp(f);
endp;


/******************************************************
SphericalBessel

Format: b=SphericalBessel(x,n)

Inputs:	
	x	rx1, points of evaluation
		nonnegative reals
	n	nonnegative integer
		order of spherical Bessel function

Output:
	b	rxm
		Spherical Bessel functions of the first kind
		Orders 1,2,...,m


********************************************************/
proc SphericalBessel(x,m);
local xi,x1,b0,b1,b2,b,mm,x2,a,xx,x22;
  x1=x+(x.<= .01);
  b0=(sin(x1)./x1);
  b1=((b0-cos(x1))./x1);
  b=b1;
  if m>1;
  @ forward recursion @
    for j (2,m,1);
      b2=-b0+(b1./x1)*(2*j-1);
      b=b~b2;
      b0=b1;b1=b2;
    endfor;
  endif;
  @ series method @
  x2=x/2;
  x22=-(x2.^2);
  for mm (1,m,1);
    for xi (1,rows(x),1);
      xx=x2[xi];
      if (xx .< (mm*.375));
        a=(xx^mm)/prodc(seqa(.5,1,mm+1))/2;
        b[xi,mm]=HyperGeometric0f1(1.5+mm,x22[xi],100,1e-8).*a;
      endif;
    endfor;
  endfor;
retp(b);
endp;


proc HyperGeometric0F1(c,x,max,tol);
local d,f,j;
  d=1; f=1;
  for j (1,max,1);
    if maxc(maxc(missrv(abs(d./f),0))) < tol; break; endif;
    d=(d.*x)./((c+j-1)*j);
    f=f+d;
  endfor;
  f=missex(f,(abs(d./f) .> tol));
retp(f);
endp;

proc HyperGeometric1f1(a,c,x,max,tol);
local d,f,j,it;
  it=0;
  d=1; f=1;
  if prodc((a.==round(a)).*(a.<0))==1;
    it=1;
    max=maxc(vec(-a));
  endif;
  for j (1,max,1);
    if maxc(maxc(missrv(abs(d./f),0))) < tol; break; endif;
    d=(d.*x).*((a+j-1)./(c+j-1)/j);
    f=f+d;
  endfor;
  if it .ne 1;
    f=missex(f,(abs(d./f) .> tol));
  endif;
retp(f);
endp;


proc HyperGeometric2f1(a,b,c,x,max,tol);
local d,f,j,it;
  it=0;
  d=1; f=1;
  if prodc((a.==round(a)).*(a.<0))==1;
    it=1;
    max=maxc(vec(-a));
  endif;
  for j (1,max,1);
    if maxc(maxc(missrv(abs(d./f),0))) < tol; break; endif;
    d=(d.*x).*((a+j-1).*(b+j-1)./(c+j-1)/j);
    f=f+d;
  endfor;
  if it .ne 1;
    f=missex(f,(abs(d./f) .> tol));
  endif;
retp(f);
endp;



